import unittest

from sqlalchemy import Integer, Unicode
from nanohttp import settings

from restfulpy.testing import WebAppTestCase
from restfulpy.testing.helpers import MockupApplication
from restfulpy.orm import DeclarativeBase, Field, DBSession, ActivationMixin, AutoActivationMixin


class ActiveObject(ActivationMixin, DeclarativeBase):
    __tablename__ = 'active_object'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50))


class AutoActiveObject(AutoActivationMixin, DeclarativeBase):
    __tablename__ = 'auto_active_object'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50))


class ActivationMixinTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', None)
    __configuration__ = '''
    db:
      url: sqlite://    # In memory DB
      echo: false
    '''

    @classmethod
    def configure_app(cls):
        cls.application.configure(force=True)
        settings.merge(cls.__configuration__)

    def test_activation_mixin(self):
        # noinspection PyArgumentList
        object1 = ActiveObject(
            title='object 1',
        )

        DBSession.add(object1)
        DBSession.commit()
        self.assertFalse(object1.is_active)
        self.assertEqual(DBSession.query(ActiveObject).filter(ActiveObject.is_active).count(), 0)

        object1.is_active = True
        self.assertTrue(object1.is_active)
        DBSession.commit()
        object1 = DBSession.query(ActiveObject).one()
        self.assertTrue(object1.is_active)

        json = object1.to_dict()
        self.assertIn('isActive', json)

        self.assertEqual(DBSession.query(ActiveObject).filter(ActiveObject.is_active).count(), 1)
        self.assertEqual(ActiveObject.filter_activated().count(), 1)

        self.assertFalse(ActiveObject.import_value(ActiveObject.is_active, 'false'))
        self.assertFalse(ActiveObject.import_value(ActiveObject.is_active, 'FALSE'))
        self.assertFalse(ActiveObject.import_value(ActiveObject.is_active, 'False'))
        self.assertTrue(ActiveObject.import_value(ActiveObject.is_active, 'true'))
        self.assertTrue(ActiveObject.import_value(ActiveObject.is_active, 'TRUE'))
        self.assertTrue(ActiveObject.import_value(ActiveObject.is_active, 'True'))
        self.assertEqual(ActiveObject.import_value(ActiveObject.title, 'title'), 'title')

    def test_auto_activation(self):
        # noinspection PyArgumentList
        object1 = AutoActiveObject(
            title='object 1',
        )

        DBSession.add(object1)
        DBSession.commit()
        self.assertTrue(object1.is_active)
        self.assertEqual(1, DBSession.query(AutoActiveObject).filter(AutoActiveObject.is_active).count())

    def test_metadata(self):
        # Metadata
        object_metadata = ActiveObject.json_metadata()
        self.assertIn('id', object_metadata['fields'])
        self.assertIn('title', object_metadata['fields'])
        self.assertIn('isActive', object_metadata['fields'])


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest
import time
from freezegun import freeze_time
from nanohttp import json, Controller, context, settings
from nanohttp.contexts import Context

from restfulpy.authentication import StatefulAuthenticator
from restfulpy.authorization import authorize
from restfulpy.principal import JwtPrincipal, JwtRefreshToken
from restfulpy.testing import WebAppTestCase, As
from restfulpy.testing.helpers import MockupApplication

session_info_test_cases = [
    {
        'environment': {
            'REMOTE_ADDR': '',
            'HTTP_USER_AGENT': 'Mozilla/5.0 (iPhone; CPU iPhone OS 5_1 like Mac OS X) AppleWebKit/534.46 (KHTML, '
                               'like Gecko) Version/5.1 Mobile/9B179 Safari/7534.48.3 RestfulpyClient-js/1.2.3 (My '
                               'App; test-name; 1.4.5-beta78; fa-IR; some; extra; info)'
        },
        'expected_remote_address': 'NA',
        'expected_machine': 'iPhone',
        'expected_os': 'iOS 5.1',
        'expected_agent': 'Mobile Safari 5.1',
        'expected_client': 'RestfulpyClient-js 1.2.3',
        'expected_app': 'My App (test-name) 1.4.5-beta78',
        'expected_last_activity': '2017-07-13T13:11:44',
    },
    {
        'environment': {
            'REMOTE_ADDR': '185.87.34.23',
            'HTTP_USER_AGENT': 'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0) '
                               'RestfulpyClient-custom/4.5.6 (A; B; C)'
        },
        'expected_remote_address': '185.87.34.23',
        'expected_machine': 'PC',
        'expected_os': 'Windows 7',
        'expected_agent': 'IE 9.0',
        'expected_client': 'RestfulpyClient-custom 4.5.6',
        'expected_app': 'A (B) C',
        'expected_last_activity': '2017-07-13T13:11:44',
    },
    {
        'environment': {
            'REMOTE_ADDR': '172.16.0.111',
            'HTTP_USER_AGENT': ''
        },
        'expected_remote_address': '172.16.0.111',
        'expected_machine': 'Other',
        'expected_os': 'Other',
        'expected_agent': 'Other',
        'expected_client': 'Unknown',
        'expected_app': 'Unknown',
        'expected_last_activity': '2017-07-13T13:11:44',
    },
    {
        'environment': {},
        'expected_remote_address': 'NA',
        'expected_machine': 'Other',
        'expected_os': 'Other',
        'expected_agent': 'Other',
        'expected_client': 'Unknown',
        'expected_app': 'Unknown',
        'expected_last_activity': '2017-07-13T13:11:44',
    }
]


class MockupMember:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


roles = ['admin', 'test']


class MockupStatefulAuthenticator(StatefulAuthenticator):
    def validate_credentials(self, credentials):
        email, password = credentials
        if password == 'test':
            return MockupMember(id=1, email=email, roles=['admin', 'test'])

    def create_refresh_principal(self, member_id=None):
        return JwtRefreshToken(dict(
            id=member_id
        ))

    def create_principal(self, member_id=None, session_id=None):
        return JwtPrincipal(dict(id=1, email='test@example.com', roles=roles, sessionId='1'))


class Root(Controller):
    @json
    def login(self):
        principal = context.application.__authenticator__.login((context.form['email'], context.form['password']))
        return dict(token=principal.dump())

    @json
    @authorize
    def me(self):
        return context.identity.payload

    @json
    @authorize
    def invalidate_token(self):
        context.application.__authenticator__.invalidate_member(1)
        return context.identity.payload

    @json(verbs='delete')
    @authorize
    def logout(self):
        context.application.__authenticator__.logout()
        return {}


class StatefulAuthenticatorTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', Root(), authenticator=MockupStatefulAuthenticator())

    @classmethod
    def configure_app(cls):
        cls.application.configure(force=True)
        settings.merge("""
            jwt:
              max_age: .3
              refresh_token:
                max_age: 3 
                secure: false               
        """)

    def test_invalidate_token(self):
        response, headers = self.request('ALL', 'GET', '/login', json=dict(email='test@example.com', password='test'))
        refresh_token = headers['Set-Cookie'].split('; ')[0]
        self.assertIn('token', response)
        self.assertTrue(refresh_token.startswith('refresh-token='))

        # Login on client
        token = response['token']
        self.wsgi_app.jwt_token = token

        # Request a protected resource to ensure authenticator is working well
        response, ___ = self.request(As.member, 'GET', '/me', headers={'Cookie': refresh_token})
        self.assertListEqual(response['roles'], roles)

        # Invalidating the token by server
        roles.append('god')
        response, headers = self.request(As.member, 'GET', '/invalidate_token', headers={'Cookie': refresh_token})
        self.assertListEqual(response['roles'], roles)
        self.assertIn('X-New-JWT-Token', headers)

        # Invalidating the token by server after the token has been expired expired, with appropriate cookies.
        time.sleep(1)
        response, headers = self.request(
            As.member, 'GET', '/invalidate_token',
            headers={
                'Cookie': refresh_token
            }
        )
        self.assertIn('X-New-JWT-Token', headers)
        self.assertIsNotNone(headers['X-New-JWT-Token'])
        self.wsgi_app.jwt_token = headers['X-New-JWT-Token']
        self.request(As.member, 'GET', '/me', headers={'Cookie': refresh_token})

    def test_logout(self):
        response, headers = self.request('ALL', 'POST', '/login', json=dict(email='test@example.com', password='test'))
        self.assertIn('token', response)
        self.assertEqual(headers['X-Identity'], '1')
        self.wsgi_app.jwt_token = response['token']
        response, headers = self.request('ALL', 'DELETE', '/logout')
        self.assertNotIn('X-Identity', headers)

    def test_session_member(self):
        with Context(environ={}, application=self.application):
            principal = self.application.__authenticator__.login(('test@example.com', 'test'))
            self.assertEqual(self.application.__authenticator__.get_session_member(principal.session_id), 1)

    @freeze_time("2017-07-13T13:11:44", tz_offset=-4)
    def test_session_info(self):
        # Login
        response, headers = self.request('ALL', 'GET', '/login', json=dict(email='test@example.com', password='test'))
        self.wsgi_app.jwt_token = response['token']

        # Testing test cases
        for test_case in session_info_test_cases:
            # Our new session info should be updated
            payload, ___ = self.request(As.member, 'GET', '/me', extra_environ=test_case['environment'])

            info = self.application.__authenticator__.get_session_info(payload['sessionId'])
            self.assertDictEqual(info, {
                'remoteAddress': test_case['expected_remote_address'],
                'machine': test_case['expected_machine'],
                'os': test_case['expected_os'],
                'agent': test_case['expected_agent'],
                'client': test_case['expected_client'],
                'app': test_case['expected_app'],
                'lastActivity': test_case['expected_last_activity'],
            })


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

from sqlalchemy import Integer, Unicode
from nanohttp import settings

from restfulpy.testing import WebAppTestCase
from restfulpy.testing.helpers import MockupApplication
from restfulpy.orm import DeclarativeBase, Field, DBSession, DeactivationMixin


class DeactiveObject(DeactivationMixin, DeclarativeBase):
    __tablename__ = 'deactive_object'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50))


class DeactivationMixinTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', None)
    __configuration__ = '''
    db:
      url: sqlite://    # In memory DB
      echo: false
    '''

    @classmethod
    def configure_app(cls):
        cls.application.configure(force=True)
        settings.merge(cls.__configuration__)

    def test_deactivation_mixin(self):
        # noinspection PyArgumentList
        object1 = DeactiveObject(
            title='object 1',
        )

        DBSession.add(object1)
        DBSession.commit()

        self.assertFalse(object1.is_active)
        self.assertEqual(DBSession.query(DeactiveObject).filter(DeactiveObject.is_active).count(), 0)

        object1.is_active = True
        self.assertTrue(object1.is_active)
        DBSession.commit()
        object1 = DBSession.query(DeactiveObject).one()
        self.assertTrue(object1.is_active)
        self.assertIsNone(object1.deactivated_at)
        self.assertIsNotNone(object1.activated_at)

        object1.is_active = False
        self.assertFalse(object1.is_active)
        DBSession.commit()
        object1 = DBSession.query(DeactiveObject).one()
        self.assertFalse(object1.is_active)
        self.assertIsNone(object1.activated_at)
        self.assertIsNotNone(object1.deactivated_at)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import argparse
import sys
import warnings
from os.path import basename
import argcomplete


from .application import Application
from .cli import Launcher


warnings.filterwarnings('ignore', message='Unknown REQUEST_METHOD')


__version__ = '0.50.0'


class RestfulpyMainLauncher(Launcher):

    def __init__(self):
        self.parser = parser = argparse.ArgumentParser(
            prog='restfulpy',
            description='Restfulpy command line interface.'
        )
        subparsers = parser.add_subparsers(
            title="Restfulpy sub commands",
            prog=basename(sys.argv[0]),
            dest="command"
        )

        from restfulpy.mockupservers import SimpleMockupServerLauncher
        SimpleMockupServerLauncher.register(subparsers)

        from restfulpy.documentary import DocumentaryLauncher
        DocumentaryLauncher.register(subparsers)

        argcomplete.autocomplete(parser)

    def launch(self, args=None):
        cli_args = self.parser.parse_args(args if args else None)
        if hasattr(cli_args, 'func'):
            cli_args.func(cli_args)
        else:
            self.parser.print_help()
        sys.exit(0)

    @classmethod
    def create_parser(cls, subparsers):
        """
        Do nothing here
        """
        pass


def main():
    RestfulpyMainLauncher()()



import re

ISO_DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'
ISO_DATETIME_PATTERN = re.compile('^(?P<datetime>\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2})(\.\d+)?Z?$')
ISO_DATE_FORMAT = '%Y-%m-%d'

POSIX_TIME_PATTERN = re.compile('^\d+(\.\d+)?$')



import base64
import ujson

from restfulpy.messaging import Messenger
from restfulpy.principal import JwtPrincipal
from restfulpy.application import Application


# noinspection PyAbstractClass
class MockupApplication(Application):
    builtin_configuration = '''
    db:
      test_url: postgresql://postgres:postgres@localhost/restfulpy_test
      administrative_url: postgresql://postgres:postgres@localhost/postgres
    logging:
      loggers:
        default:
          level: critical
    '''

    def configure(self, files=None, context=None, **kwargs):
        _context = dict(
            process_name='restfulpy_unittests'
        )
        if context:
            _context.update(context)
        super().configure(files=files, context=_context, **kwargs)

class MockupMessenger(Messenger):
    _last_message = None

    @property
    def last_message(self):
        return self.__class__._last_message

    @last_message.setter
    def last_message(self, value):
        self.__class__._last_message = value

    def send(
            self,
            to, subject, body,
            cc=None,
            bcc=None,
            template_string=None,
            template_filename=None,
            from_=None,
            attachments=None
    ):
        self.last_message = {
            'to': to,
            'body': body,
            'subject': subject
        }


class UnsafePrincipal(JwtPrincipal):  # pragma: no cover
    @classmethod
    def load(cls, encoded, force=False):
        decoded = base64.b64decode('%s=' % encoded.split('.')[1])  # To avoid padding exception
        payload = ujson.loads(decoded)
        return cls(payload)



import time
import unittest
from nanohttp import json, Controller, context, settings
from restfulpy.authentication import StatefulAuthenticator
from restfulpy.authorization import authorize
from restfulpy.principal import JwtPrincipal, JwtRefreshToken
from restfulpy.testing import WebAppTestCase, As
from restfulpy.testing.helpers import MockupApplication


class MockupMember:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


roles = ['admin', 'test']


class MockupAuthenticator(StatefulAuthenticator):
    def validate_credentials(self, credentials):
        email, password = credentials
        if password == 'test':
            return MockupMember(id=1, email=email, roles=['admin', 'test'])

    def create_refresh_principal(self, member_id=None):
        return JwtRefreshToken(dict(
            id=member_id
        ))

    def create_principal(self, member_id=None, session_id=None):
        return JwtPrincipal(dict(id=1, email='test@example.com', roles=roles, sessionId='1'))


class Root(Controller):

    @json
    def login(self):
        principal = context.application.__authenticator__.login(
            context.form['email'], context.form['password']
        )

        return dict(token=principal.dump())


    @json
    @authorize
    def me(self):
        return context.identity.payload

    @json
    @authorize
    def invalidate_token(self):
        context.application.__authenticator__.invalidate_member(1)
        return context.identity.payload


class StatefulAuthenticatorTestCase(WebAppTestCase):
    application = MockupApplication(
        'MockupApplication', Root(), authenticator=MockupAuthenticator())

    @classmethod
    def configure_app(cls):
        cls.application.configure(force=True)
        settings.merge("""
            jwt:
              max_age: .3
              refresh_token:
                max_age: 3 
                secure: true          
        """)

    def test_invalidate_token(self):
        principal = JwtPrincipal(dict(
            email='test@example.com',
            id=1,
            sessionId=1,
            roles=roles
        ))

        token = principal.dump().decode("utf-8")
        refresh_token = 'refresh-token=' + JwtRefreshToken(dict(id=1)).dump().decode("utf-8")
        self.assertTrue(refresh_token.startswith('refresh-token='))
        self.wsgi_app.jwt_token = token

        time.sleep(1)
        self.request(As.member, 'GET', '/me', headers={'Cookie': refresh_token}, expected_status=400)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest
from urllib.parse import urlencode
from os import path

import webtest
from nanohttp import settings
import ujson

from restfulpy.db import DatabaseManager
from restfulpy.orm import setup_schema, session_factory, create_engine
from restfulpy.application import Application


class WSGIDocumentaryTestCase(unittest.TestCase):
    application = None
    api_client = None
    controller_factory = None
    application_factory = None
    fields = None

    @staticmethod
    def documentary_middleware_factory(app):
        return FileDocumentaryMiddleware(app, path.abspath(path.join(settings.api_documents.directory, 'source')))

    @classmethod
    def application_factory(cls):
        if cls.controller_factory:
            app = Application(f'{cls.__name__}Application', cls.controller_factory())
            app.configure(force=True)
            return app
        else:
            raise ValueError('One of controller_factory and or application_factory must be provided.')

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.application = cls.application_factory()
        cls.api_client = webtest.TestApp(cls.documentary_middleware_factory(cls.application))

    def call(self, title, verb, url, *, query=None, environ=None, description=None, form=None, role=None, status=200,
             **kwargs):

        environ = environ or {}
        environ['TEST_CASE_TITLE'] = title

        if description:
            environ['TEST_CASE_DESCRIPTION'] = description

        if role:
            environ['TEST_CASE_ROLE'] = role

        if status:
            environ['TEST_CASE_EXPECTED_STATUS'] = str(status)

        if self.fields:
            environ['TEST_CASE_FIELDS'] = ujson.dumps({n: f.to_dict() for n, f in self.fields.items()})

        if query:
            if isinstance(query, str):
                url += f'?{query}'
            else:
                url = f'{url}?{urlencode(query, doseq=True)}'

        if form:
            kwargs['params'] = form

        response = self.api_client._gen_request(
            verb.upper(), url,
            expect_errors=status is not None,
            extra_environ=environ,
            **kwargs
        )
        if status:
            self.assertEqual(status, response.status_code)
        return response


# noinspection PyAbstractClass
class RestfulpyApplicationTestCase(WSGIDocumentaryTestCase):
    session = None
    engine = None

    @classmethod
    def application_factory(cls):
        app = super().application_factory()
        app.configure(force=True, context=dict(unittest=True))
        settings.merge('''
            messaging:
              default_messenger: restfulpy.testing.MockupMessenger
            logging:
              loggers:
                default:
                  level: critical
        ''')
        settings.db.url = settings.db.test_url
        return app

    @classmethod
    def prepare_database(cls):
        with DatabaseManager() as m:
            m.drop_database()
            m.create_database()

        cls.engine = create_engine()
        cls.session = session = session_factory(bind=cls.engine, expire_on_commit=False)
        setup_schema(session)
        session.commit()

    @classmethod
    def drop_database(cls):
        cls.session.close_all()
        cls.engine.dispose()
        with DatabaseManager() as m:
            m.drop_database()

    @classmethod
    def mockup(cls):
        pass

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.prepare_database()
        cls.application.initialize_models()
        cls.mockup()

    @classmethod
    def tearDownClass(cls):
        cls.application.shutdown()
        cls.drop_database()
        super().tearDownClass()



import collections
from os import path, makedirs

from .call import ApiCall


class AbstractDocumentaryMiddleware:

    def __init__(self, application, history_maxlen=50):
        self.application = application
        self.call_history = collections.deque(maxlen=history_maxlen)

    def on_call_done(self, call):
        raise NotImplementedError()

    def __call__(self, environ, start_response):
        call = ApiCall.from_environ(environ)

        response = self.application(
            environ,
            call.start_response_wrapper(start_response)
        )
        call.response.load(response)

        if not call.title.startswith('SKIP:'):
            self.call_history.append(call)
            self.on_call_done(call)
        for i in call.response.buffer:
            yield i


class FileDocumentaryMiddleware(AbstractDocumentaryMiddleware):

    def __init__(self, application, directory=None):
        super().__init__(application)
        directory = directory.rstrip('/')
        if not path.exists(directory):
            makedirs(directory, exist_ok=True)
        self.directory = directory

    def on_call_done(self, call):
        call.save(self.directory)



import io
import re
from urllib.parse import parse_qs
from os import path

import ujson
import yaml
from nanohttp.helpers import parse_any_form

from .constants import  URL_PARAMETER_PATTERN, CONTENT_TYPE_PATTERN
from ..fieldinfo import FieldInfo


class Response:
    _status = None
    status_code = None
    status_text = None
    content_type = None
    encoding = None

    def __init__(self, status=None, headers=None, body=None, status_code=None, status_text=None):
        self.status_code = status_code
        self.status_text = status_text
        if status:
            self.status = status

        if isinstance(body, dict):
            body = ujson.encode(body)

        self.buffer = body.splitlines() if body else []
        self.headers = headers or []

        for i in self.headers:
            k, v = i.split(': ') if isinstance(i, str) else i
            if k == 'Content-Type':
                match = CONTENT_TYPE_PATTERN.match(v)
                if match:
                    self.content_type, self.encoding = match.groups()
                break

    def load(self, wsgi_response):
        for chunk in wsgi_response:
            self.buffer.append(chunk)
        return self.buffer

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        self._status = value
        if ' ' in value:
            parts = value.split(' ')
            status_code, self.status_text = parts[0], ' '.join(parts[1:])
        else:
            status_code = value
        self.status_code = int(status_code)

    def to_dict(self):
        return dict(
            status_code=self.status_code,
            status_text=self.status_text,
            headers=[f'{k}: {v}' for k, v in self.headers],
            body=self.json if self.content_type == 'application/json' else self.text
        )

    @property
    def body(self):
        payload = b''.join(self.buffer)
        return payload

    @property
    def text(self):
        return self.body.decode()

    @property
    def json(self):
        return ujson.loads(self.body)


class ApiCall:
    response = None
    url_parameters = None
    query_string = None

    def __init__(self, url, verb='GET', title=None, description=None, query_string=None,
                 form=None, role=None, expected_status=None, url_parameters=None, response=None):
        self.description = description
        self.query_string = query_string
        self.form = form
        self.url = url
        self.title = title
        self.verb = verb
        self.role = role
        self.expected_status = expected_status
        self.url_parameters = url_parameters
        if response:
            self.response = response if isinstance(response, Response) else Response(**response)

    @classmethod
    def from_environ(cls, environ, *args, **kwargs):
        if environ['QUERY_STRING']:
            kwargs['query_string'] = {
                k: v[0] if len(v) == 1 else v for k, v in parse_qs(
                    environ['QUERY_STRING'],
                    keep_blank_values=True,
                    strict_parsing=False
                ).items()
            }

        kwargs['description'] = environ.get('TEST_CASE_DESCRIPTION', None)
        kwargs['title'] = environ['TEST_CASE_TITLE']
        kwargs['verb'] = environ['REQUEST_METHOD'].upper()
        kwargs['role'] = environ.get('TEST_CASE_ROLE', None)

        status = environ.get('TEST_CASE_EXPECTED_STATUS', None)
        kwargs['expected_status'] = int(status) if status else status

        url = environ['PATH_INFO']
        if URL_PARAMETER_PATTERN.search(url):
            # noinspection PyTypeChecker
            url_parameters = dict()
            for k, v in URL_PARAMETER_PATTERN.findall(url):
                url_parameters[k] = v
                url = re.sub(f'{k}:\s?', '', url)
            environ['PATH_INFO'] = url
        else:
            url_parameters = None
        kwargs['url_parameters'] = url_parameters

        fields = environ.get('TEST_CASE_FIELDS')
        fields = ujson.loads(fields) if fields else {}
        kwargs['form'] = cls.parse_form(environ, fields=fields)
        return cls(url, *args, **kwargs)

    @staticmethod
    def parse_form(environ, fields):
        form_file = environ['wsgi.input']
        content_length = int(environ.get('CONTENT_LENGTH', 0))
        form_data = form_file.read(content_length)
        monkey_form_file = io.BytesIO(form_data)
        environ['wsgi.input'] = monkey_form_file
        form = parse_any_form(environ)
        monkey_form_file.seek(0)
        return {k: ApiField(v, **fields.get(k, {})) for k, v in form.items()} if form else None

    def start_response_wrapper(self, start_response):
        def start_response_profiler(status, headers):
            self.response = Response(status, headers)
            return start_response(status, headers)
        return start_response_profiler

    def to_dict(self):
        return dict(
            title=self.title,
            url=self.url,
            url_parameters=self.url_parameters,
            verb=self.verb,
            query_string=self.query_string,
            response=self.response.to_dict(),
            form={k: v.to_dict() for k, v in self.form.items()} if self.form else None,
            role=self.role,
            expected_status=self.expected_status,
            description=self.description
        )

    def dump(self, file):
        yaml.dump(self.to_dict(), file, default_flow_style=False)

    def save(self, directory):
        with open(path.join(directory, self.filename), 'w') as f:
            self.dump(f)

    @property
    def filename(self):
        url = self.url.strip('/').replace('/', '-')
        title = self.title.replace(' ', '-')
        return f'{url}-{self.verb}-{self.response.status}-{title}.yml'.replace(' ', '-')

    @classmethod
    def load(cls, file):
        data = yaml.load(file)
        return cls(**data)


class ApiField(FieldInfo):
    def __init__(self, value, **kwargs):
        super().__init__(**kwargs)
        import builtins
        t = getattr(builtins, self.type_, str)
        self.value = t(value)




import unittest

from nanohttp import Controller, text, context, json

from restfulpy.fieldinfo import FieldInfo
from restfulpy.documentary import AbstractDocumentaryMiddleware, WSGIDocumentaryTestCase, Response, ApiField

last_call = None


class ExaminationMiddleware(AbstractDocumentaryMiddleware):
    def on_call_done(self, call):
        global last_call
        last_call = call


class Root(Controller):
    @text
    def index(self, id_='empty'):
        yield f'Content {id_}'

    @text
    def books(self, isbn, name=None):
        yield f'Book {isbn}'
        if name:
            yield f' {name}'

    @text
    def search(self, resource, *, query=None):
        yield f'{resource} {query}'

    @json
    def signup(self):
        return context.form


class DocumentaryTestCase(WSGIDocumentaryTestCase):
    documentary_middleware_factory = ExaminationMiddleware
    controller_factory = Root

    fields = {
        'field1': FieldInfo(type_=int),
        'field2': FieldInfo(type_=str)
    }

    def test_basic_pipeline(self):
        response = self.call('Simple pipeline', 'GET', '/')
        self.assertEqual('Content empty', response.text)

        self.assertEqual('Simple pipeline', last_call.title)
        self.assertEqual('/', last_call.url)
        self.assertIsNone(last_call.url_parameters)
        self.assertIsNone(last_call.query_string)
        self.assertIsInstance(last_call.response, Response)
        self.assertEqual('200 OK', last_call.response.status)
        self.assertEqual(200, last_call.response.status_code)
        self.assertEqual('OK', last_call.response.status_text)
        self.assertEqual(b'Content empty', last_call.response.body)
        self.assertEqual('Content empty', last_call.response.text)
        self.assertEqual('text/plain', last_call.response.content_type)
        self.assertEqual('utf-8', last_call.response.encoding)

        self.assertDictEqual(last_call.to_dict(), dict(
            title='Simple pipeline',
            url='/',
            url_parameters=None,
            verb='GET',
            query_string=None,
            form=None,
            role=None,
            expected_status=200,
            description=None,
            response=dict(
                body='Content empty',
                headers=['Content-Type: text/plain; charset=utf-8'],
                status_code=200,
                status_text='OK'
            )
        ))

    def test_skip(self):
        global last_call
        last_call = None
        response = self.call('SKIP:', 'GET', '/')
        self.assertEqual('Content empty', response.text)
        self.assertIsNone(last_call)

    def test_url_parameters(self):
        response = self.call('Url parameters 1', 'GET', '/id: 1')
        self.assertEqual('Content 1', response.text)
        self.assertEqual('/1', last_call.url)
        self.assertDictEqual(dict(id='1'), last_call.url_parameters)

        response = self.call('Url parameters 2', 'GET', '/books/isbn: 2')
        self.assertEqual('Book 2', response.text)
        self.assertEqual('/books/2', last_call.url)
        self.assertDictEqual(dict(isbn='2'), last_call.url_parameters)

        response = self.call('Url parameters 3', 'GET', '/books/isbn: 2/name: abc')
        self.assertEqual('Book 2 abc', response.text)
        self.assertEqual('/books/2/abc', last_call.url)
        self.assertDictEqual(dict(isbn='2', name='abc'), last_call.url_parameters)

    def test_query_string(self):
        response = self.call('Query string', 'GET', '/search/books', query=dict(query='abc'))
        self.assertEqual('books abc', response.text)
        self.assertDictEqual(dict(query='abc'), last_call.query_string)

    def test_forms(self):
        response = self.call(
            'Url-encoded form',
            'POST',
            '/signup',
            form=dict(
                field1=12,
                field2='two'
            )
        )
        expected_dict = dict(field1='12', field2='two')
        self.assertDictEqual(expected_dict, response.json)
        self.assertDictEqual(
            ApiField(type_=int, value=12).to_dict(),
            last_call.form['field1'].to_dict()
        )
        self.assertDictEqual(dict(
                type_='int',
                value=12,
                default=None,
                optional=False,
                pattern=None,
                max_length=None,
                min_length=None,
                readonly=False,
                protected=False
            ),
            last_call.to_dict()['form']['field1']
        )


if __name__ == '__main__':  # pragma: no cover
    unittest.main()




import time
import argparse
import threading
from os import remove
import tempfile
from os.path import dirname, abspath, join, exists
from subprocess import run
from wsgiref.simple_server import make_server

from sqlalchemy import Integer, Unicode
from nanohttp import text, json, context, RestController, HttpBadRequest, etag, HttpNotFound
from restfulpy.authorization import authorize
from restfulpy.application import Application
from restfulpy.authentication import StatefulAuthenticator
from restfulpy.controllers import RootController, ModelRestController, JsonPatchControllerMixin
from restfulpy.orm import DeclarativeBase, OrderingMixin, PaginationMixin, FilteringMixin, Field, setup_schema, \
    DBSession, ModifiedMixin, commit
from restfulpy.principal import JwtPrincipal, JwtRefreshToken
from restfulpy.cli import Launcher
from restfulpy.templating import template


__version__ = '0.1.1'


here = abspath(dirname(__file__))
db = abspath(join(tempfile.gettempdir(), 'restfulpy-mockup-server.sqlite'))


class MockupMember:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class Resource(ModifiedMixin, OrderingMixin, PaginationMixin, FilteringMixin, DeclarativeBase):
    __tablename__ = 'resource'

    id = Field(Integer, primary_key=True)
    title = Field(
        Unicode(30),
        watermark='title here',
        icon='default',
        label='Title',
        pattern='[a-zA-Z0-9]{3,}',
        min_length=3
    )

    @property
    def __etag__(self):
        return self.last_modification_time.isoformat()


class MockupAuthenticator(StatefulAuthenticator):
    def validate_credentials(self, credentials):
        email, password = credentials
        if email == 'user1@example.com' and password == '123456':
            return MockupMember(id=1, email=email, roles=['user'])

    def create_refresh_principal(self, member_id=None):
        return JwtRefreshToken(dict(id=member_id))

    def create_principal(self, member_id=None, session_id=None):
        return JwtPrincipal(dict(id=1, email='user1@example.com', roles=['user'], sessionId='1'))


class AuthController(RestController):

    @json
    def post(self):
        email = context.form.get('email')
        password = context.form.get('password')

        def bad():
            raise HttpBadRequest('Invalid email or password')

        if not (email and password):
            bad()

        principal = context.application.__authenticator__.login((email, password))
        if principal is None:
            bad()

        return dict(token=principal.dump())

    @json
    def delete(self):
        return {}


class ResourceController(JsonPatchControllerMixin, ModelRestController):
    __model__ = Resource

    @json
    @etag
    @Resource.expose
    def get(self, id_: int=None):
        q = Resource.query
        if id_ is not None:
            return q.filter(Resource.id == id_).one_or_none()
        return q

    @json
    @etag
    @Resource.expose
    @commit
    def put(self, id_: int=None):
        m = Resource.query.filter(Resource.id == id_).one_or_none()
        if m is None:
            raise HttpNotFound()
        m.update_from_request()
        context.etag_match(m.__etag__)
        return m

    @json
    @etag
    @commit
    def post(self):
        m = Resource()
        m.update_from_request()
        DBSession.add(m)
        return m

    @json
    @etag
    @commit
    def delete(self, id_: int=None):
        m = Resource.query.filter(Resource.id == id_).one_or_none()
        context.etag_match(m.__etag__)
        DBSession.delete(m)
        return m


class Root(RootController):
    resources = ResourceController()
    sessions = AuthController()

    @template('help.mako')
    def index(self):
        return dict(url=f'http://{context.environ["HTTP_HOST"]}')

    @json
    def echo(self):
        return {k: v for i in (context.form, context.query_string) for k, v in i.items()}

    @text
    @authorize
    def protected(self):
        return 'Protected'

    @json
    def version(self):
        return {
            'version': __version__
        }


class MockupApplication(Application):
    __authenticator__ = MockupAuthenticator()
    builtin_configuration = f'''
    debug: true
    
    db:
      url: sqlite:///{db}
      
    jwt:
      max_age: 20
      refresh_token:
        max_age: 60
        secure: false
    
    templates:
      directories:
        - {here}/templates    
    '''

    def __init__(self):
        super().__init__(
            'restfulpy-client-js-mockup-server',
            root=Root(),
            version=__version__,
        )

    def insert_basedata(self):
        pass

    def insert_mockup(self):
        for i in range(1, 11):
            # noinspection PyArgumentList
            DBSession.add(Resource(id=i, title='resource%s' % i))
        DBSession.commit()

    def begin_request(self):
        if 'HTTP_ORIGIN' in context.environ:
            context.response_headers.add_header('Access-Control-Allow-Origin', context.environ['HTTP_ORIGIN'])
        super(MockupApplication, self).begin_request()


class SimpleMockupServerLauncher(Launcher):
    __command__ = 'mockup-server'
    default_bind = '8080'
    application = None

    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser(cls.__command__, help='Starts the mockup http server.')
        parser.add_argument(
            '-c', '--config-file',
            metavar="FILE",
            help='List of configuration files separated by space. Default: ""'
        )
        parser.add_argument(
            '-b', '--bind',
            default=cls.default_bind,
            metavar='{HOST:}PORT',
            help=f'Bind Address. default is {cls.default_bind}, A free tcp port will be choosed automatically if the '
                 f'0 (zero) is given'
        )

        parser.add_argument(
            'command',
            nargs=argparse.REMAINDER,
            default=[],
            help='The command to run tests.'
        )

        parser.add_argument('-v', '--version', action='store_true', help='Show the mockup server\'s version.')
        return parser

    def launch(self):
        if self.args.version:
            print(__version__)
            return

        self.application = MockupApplication()
        self.application.configure(files=self.args.config_file)
        if exists(db):
            remove(db)
        # makedirs(settings.data_directory, exist_ok=True)
        self.application.initialize_models()
        setup_schema()
        # DBSession.commit()
        print(f'DB {DBSession.bind}')
        self.application.insert_mockup()
        host, port = self.args.bind.split(':') if ':' in self.args.bind else ('localhost', self.args.bind)
        httpd = make_server(host, int(port), self.application)

        url = 'http://%s:%d' % httpd.server_address
        print(f'The server is up!, Get {url} to more info about how to use it!')
        server_thread = threading.Thread(target=httpd.serve_forever, daemon=True)
        try:
            server_thread.start()

            if not self.args.command:
                server_thread.join()
            else:
                test_runner_command = ' '.join(self.args.command).replace('{url}', url)
                time.sleep(1)
                run(test_runner_command, shell=True)
            return 0
        except KeyboardInterrupt:
            print('CTRL+X is pressed.')
            return 1
        finally:
            httpd.shutdown()
            server_thread.join()



import unittest

from restfulpy.cryptography import AESCipher


def mockup_random(size):
    return b'#' * size


class AESCipherTestCase(unittest.TestCase):
    def test_encrypt(self):
        key = b'1234567890123456'
        message = b'test-message'
        encrypted = b'################>T2\tyZ\\\x0f20[\x97e\x97\xbb5'
        cipher = AESCipher(key, random=mockup_random)
        self.assertEqual(encrypted, cipher.encrypt(message))
        self.assertEqual(message, cipher.decrypt(encrypted))


if __name__ == '__main__':  # pragma: no cover
    unittest.main()




import argparse
import sys

import argcomplete

from ...cli import Launcher
from .database import DatabaseLauncher
from .migrate import MigrateLauncher
from .serve import ServeLauncher
from .worker import WorkerLauncher
from .dev import DevLauncher
from .configuration import ConfigurationLauncher
from .jwttoken import JWTLauncher


class MainLauncher(Launcher):

    def __init__(self, application):
        self.application = application
        self.parser = parser = argparse.ArgumentParser(
            prog=application.name,
            description='%s command line interface.' % application.name
        )
        parser.add_argument(
            '-p', '--process-name',
            metavar="PREFIX",
            default=application.name,
            help='A string indicates the logger prefix for this process, it helps to configure '
                 'separate log files per process.'
        )
        parser.add_argument(
            '-c', '--config-file',
            metavar="FILE",
            help='List of configuration files separated by space. Default: ""'
        )
        parser.add_argument(
            '-d', '--config-dir',
            metavar="DIR",
            help='List of configuration directories separated by space. Default: ""'
        )
        subparsers = parser.add_subparsers(
            title="sub commands",
            prog=application.name,
            dest="command"
        )

        DatabaseLauncher.register(subparsers)
        ServeLauncher.register(subparsers)
        MigrateLauncher.register(subparsers)
        WorkerLauncher.register(subparsers)
        DevLauncher.register(subparsers)

        if application.__configuration_cipher__ is not None:
            ConfigurationLauncher.register(subparsers)

        if application.__authenticator__ is not None:
            JWTLauncher.register(subparsers)

        application.register_cli_launchers(subparsers)
        argcomplete.autocomplete(parser)

    def launch(self, args=None):
        cli_args = self.parser.parse_args(args)
        cli_args.application = self.application
        self.application.process_name = cli_args.process_name
        self.application.configure(files=cli_args.config_file, directories=cli_args.config_dir)
        self.application.initialize_models()
        if hasattr(cli_args, 'func'):
            cli_args.func(cli_args)
        else:
            self.parser.print_help()
        sys.exit(0)

    @classmethod
    def create_parser(cls, subparsers):
        """
        Do nothing here
        """
        pass



class FieldInfo:
    def __init__(self, type_=str, default=None, optional=False, pattern=None, max_length=None, min_length=None,
                 readonly=False, protected=False):
        self.type_ = type_ if isinstance(type_, str) else type_.__name__
        self.default = default
        self.optional = optional
        self.pattern = pattern
        self.max_length = max_length
        self.min_length = min_length
        self.readonly = readonly
        self.protected = protected

    def to_json(self):
        return dict(
            type_=self.type_,
            default_=self.default,
            optional=self.optional,
            pattern=self.pattern,
            maxLength=self.max_length,
            minLength=self.min_length,
            readonly=self.readonly,
            protected=self.protected
        )

    def __copy__(self):
        new_one = type(self)()
        new_one.__dict__.update(self.__dict__)
        return new_one

    def to_dict(self):
        return self.__dict__




from .call import ApiCall, Response, ApiField
from .constants import CONTENT_TYPE_PATTERN, URL_PARAMETER_PATTERN
from .middlewares import AbstractDocumentaryMiddleware, FileDocumentaryMiddleware
from .testcases import WSGIDocumentaryTestCase, RestfulpyApplicationTestCase
from .launchers import DocumentaryLauncher


from glob import glob
from os.path import join

from .call import ApiCall


class DocumentFormatter:

    def __init__(self, output_directory):
        self.output_directory = output_directory
        self.locations = {}

    @property
    def output_format(self):
        raise NotImplementedError()

    def load(self, directory):
        for filename in glob(join(directory, '*.yml')):
            with open(filename) as f:
                self.load_file(f)
        print(self.locations)

    def load_file(self, f):
        call = ApiCall.load(f)
        verbs = self.locations.setdefault(call.url, {})
        calls = verbs.setdefault(call.verb, [])
        calls.append(call)

    def dump(self):
        for location, verbs in self.locations.items():
            for verb, calls in verbs.items():
                self.write_file(self.get_filename(location, verb), calls)

    def get_filename(self, location, verb):
        url = location.strip('/').replace('/', '-')
        return f'{url}-{verb}.{self.output_format}'.replace(' ', '-')

    def write_file(self, filename, calls):
        with open(filename) as file:
            raise NotImplementedError()


class MarkdownFormatter(DocumentFormatter):
    pass



import copy
import unittest

from nanohttp import context, json, settings

from restfulpy.principal import DummyIdentity
from restfulpy.testing import WebAppTestCase
from restfulpy.testing.helpers import MockupApplication
from restfulpy.validation import validate_form
from restfulpy.controllers import RestController, RootController


class ValidationRequiresController(RestController):

    @json
    @validate_form(requires=['requiresParamForAll'],
                   client={'requires': ['requiresParamForClient']},
                   admin={'requires': ['requiresParamForAdmin']})
    def post(self):
        result = copy.deepcopy(context.form)
        result.update(context.query_string)
        return result


class Root(RootController):
    validation = ValidationRequiresController()


class ValidationRequiresTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', Root())

    @classmethod
    def configure_app(cls):
        super().configure_app()
        settings.merge("""
            logging:
              loggers:
                default:
                  level: info
            """)

    def test_validation_requires(self):
        # Test `requires`
        # role -> All
        self.wsgi_app.jwt_token = DummyIdentity().dump().decode()

        result, ___ = self.request('All', 'POST', '/validation', doc=False,
                                   params={'requiresParamForAll': 'param'})
        self.assertIn('requiresParamForAll', result)

        result, ___ = self.request(
            'All', 'POST', '/validation', doc=False,
            params={
                'customParam': 'param',
                'requiresParamForAll': 'param',
            }
        )
        self.assertIn('customParam', result)
        self.assertIn('requiresParamForAll', result)

        self.request('All', 'POST', '/validation', doc=False, expected_status=400)

        self.request('All', 'POST', '/validation', doc=False, params={'customParam': 'param'},
                     expected_status=400)

        self.request('All', 'POST', '/validation', doc=False, params={'requiresParamForClient': 'param'},
                     expected_status=400)

        # -----------------------------
        # role -> Client
        self.wsgi_app.jwt_token = DummyIdentity('client').dump().decode()
        result, ___ = self.request(
            'Client', 'POST', '/validation', doc=False,
            params={
                'requiresParamForAll': 'param',
                'requiresParamForClient': 'param',
            }
        )
        self.assertIn('requiresParamForAll', result)
        self.assertIn('requiresParamForClient', result)

        result, ___ = self.request(
            'Client', 'POST', '/validation', doc=False,
            params={
                'customParam': 'param',
                'requiresParamForAll': 'param',
                'requiresParamForClient': 'param',
            }
        )
        self.assertIn('customParam', result)
        self.assertIn('requiresParamForAll', result)
        self.assertIn('requiresParamForClient', result)

        self.request('Client', 'POST', '/validation', doc=False, expected_status=400)

        self.request('Client', 'POST', '/validation', doc=False, params={'customParam': 'param'},
                     expected_status=400)

        self.request('Client', 'POST', '/validation', doc=False, params={'requiresParamForAll': 'param'},
                     expected_status=400)

        self.request('Client', 'POST', '/validation', doc=False, params={
            'requiresParamForClient':
                'param',
        }, expected_status=400)
        # -----------------------------
        # role -> Admin
        self.wsgi_app.jwt_token = DummyIdentity('admin').dump().decode()
        result, ___ = self.request('Admin', 'POST', '/validation', doc=False, params={
            'requiresParamForAll': 'param',
            'requiresParamForAdmin': 'param',
        })
        self.assertIn('requiresParamForAll', result)
        self.assertIn('requiresParamForAdmin', result)
        result, ___ = self.request(
            'Admin', 'POST', '/validation', doc=False,
            params={
                'customParam': 'param',
                'requiresParamForAll': 'param',
                'requiresParamForAdmin': 'param',
            }
        )
        self.assertIn('customParam', result)
        self.assertIn('requiresParamForAll', result)
        self.assertIn('requiresParamForAdmin', result)

        self.request('Admin', 'POST', '/validation', doc=False, expected_status=400)

        self.request('Admin', 'POST', '/validation', doc=False, params={'customParam': 'param'},
                     expected_status=400)

        self.request(
            'Admin', 'POST', '/validation', doc=False,
            params={
                'requiresParamForAll': 'param',
                'requiresParamForClient': 'param',
            },
            expected_status=400
        )

        self.request('Admin', 'POST', '/validation', doc=False, params={'requiresParamForAdmin': 'param'},
                     expected_status=400)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()




from ..cli import Launcher, RequireSubCommand


class DocumentaryLauncher(Launcher, RequireSubCommand):
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser('documentary', help='Documentary tools')
        documentary_subparsers = parser.add_subparsers(title='documentary command', dest='documentary_command')
        DocumentGeneratorLauncher.register(documentary_subparsers)
        return parser


class DocumentGeneratorLauncher(Launcher):
    __command__ = 'generate'

    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser(cls.__command__, help='Generates the documentation from yaml test specs')
        parser.add_argument('output-directory', help='Path to the directory to store generated files')
        parser.add_argument(
            '-d', '--input-directory',
            default='.',
            help='Path to the directory which contains the test spec yaml(*.yml) files. '
                 'if omitted, the current directory will be used'
        )
        parser.add_argument(
            '-f', '--format',
            default='md',
            help='The output format, supported formats: md'
        )
        return parser

    def launch(self):
        from .formatters import MarkdownFormatter
        formatter = MarkdownFormatter()
        formatter.load(self.args.input_directory)
        formatter.dump(self.args.output_directory)





import re

from sqlalchemy import Column, Unicode, String
from sqlalchemy.orm import relationship as sa_relationship, composite as sa_composite
from nanohttp import HttpBadRequest

from ..utils import to_camel_case
from ..fieldinfo import FieldInfo


# noinspection PyAbstractClass
class Field(Column):

    @property
    def can_validate(self):
        return 'pattern' in self.info or \
            'min_length' in self.info or \
            'max_length' in self.info

    def __init__(self,
                 *args,
                 json=None,
                 readonly=None,
                 max_length=None,
                 min_length=None,
                 pattern=None,
                 protected=None,
                 watermark=None,
                 nullable=False,
                 label=None,
                 icon=None,
                 example=None,
                 message=None,
                 info=None,
                 **kwargs):
        info = info or dict()

        if json is not None:
            info['json'] = json

        if readonly is not None:
            info['readonly'] = readonly

        if protected is not None:
            info['protected'] = protected

        if watermark is not None:
            info['watermark'] = watermark

        if max_length is not None:
            info['max_length'] = max_length
        elif args and isinstance(args[0], (Unicode, String)):
            info['max_length'] = args[0].length

        if min_length is not None:
            info['min_length'] = min_length

        if pattern is not None:
            info['pattern'] = pattern

        if label is not None:
            info['label'] = label

        if icon is not None:
            info['icon'] = icon

        if example is not None:
            info['example'] = example

        if message is not None:
            info['message'] = message

        super(Field, self).__init__(*args, info=info, nullable=nullable, **kwargs)

    def _validate_pattern(self, value):
        if value is None:
            return
        if not re.match(self.info['pattern'], value):
            raise HttpBadRequest('Cannot match field: %s with value "%s" by acceptable pattern' % (self.name, value))
        return value

    def _validate_length(self, value, min_length, max_length):
        if value is None:
            return

        if not isinstance(value, str):
            raise HttpBadRequest(status_text='Invalid type: %s for field: %s' % (type(value), self.name))

        value_length = len(value)
        if min_length is not None:
            if value_length < min_length:
                raise HttpBadRequest(
                    status_text=f'Please enter at least {min_length} characters for field: {self.name}.'
                )

        if max_length is not None:
            if value_length > max_length:
                raise HttpBadRequest(
                    status_text=f'Cannot enter more than: {max_length} in field: {self.name}.'
                )

    def validate(self, value):
        if 'pattern' in self.info:
            self._validate_pattern(value)

        if 'min_length' in self.info or 'max_length' in self.info:
            self._validate_length(value, self.info.get('min_length'), self.info.get('max_length'))
        return value


def relationship(*args, json=None, protected=None, **kwargs):
    info = dict()

    if json:
        info['json'] = json

    if protected:
        info['protected'] = protected

    return sa_relationship(*args, info=info, **kwargs)


def composite(*args, json=None, protected=None, readonly=None, **kwargs):
    info = dict()

    if json:
        info['json'] = json

    if protected:
        info['protected'] = protected

    if readonly is not None:
        info['readonly'] = readonly

    return sa_composite(*args, info=info, **kwargs)


class ModelFieldInfo(FieldInfo):
    def __init__(self, name, key, primary_key=False, label=None, watermark=None, icon=None,
                 message='Invalid value', example=None, **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.key = key[1:] if key.startswith('_') else key
        self.primary_key = primary_key
        self.label = label or watermark
        self.watermark = watermark
        self.icon = icon
        self.message = message
        self.example = example

    @classmethod
    def from_column(cls, c, info=None):
        if not info:
            info = c.info
        json_name = info.get('json', to_camel_case(c.key))
        result = []

        key = c.key

        if hasattr(c, 'default') and c.default:
            default_ = c.default.arg if c.default.is_scalar else None
        else:
            default_ = None

        if hasattr(c, 'type'):
            try:
                type_ = c.type.python_type
            except NotImplementedError:
                # As we spoke, hybrid properties have no type
                type_ = ''
        else:  # pragma: no cover
            type_ = 'str'

        result.append(cls(
            json_name,
            key,
            type_=type_,
            default=default_,
            optional=c.nullable if hasattr(c, 'nullable') else None,
            pattern=info.get('pattern'),
            max_length=info.get('max_length') if 'max_length' in info else
                (c.type.length if hasattr(c, 'type') and hasattr(c.type, 'length') else None),
            min_length=info.get('min_length'),
            message=info.get('message', 'Invalid Value'),
            watermark=info.get('watermark', None),
            label=info.get('label', None),
            icon=info.get('icon', None),
            example=info.get('example', None),
            primary_key=hasattr(c.expression, 'primary_key') and c.expression.primary_key,
            readonly=info.get('readonly', False),
            protected=info.get('protected', False)
        ))

        return result

    def to_json(self):
        result = super().to_json()
        result.update(
            name=self.name,
            example=self.example,
            message=self.message,
            watermark=self.watermark,
            label=self.label,
            icon=self.icon,
            key=self.key,
            primaryKey=self.primary_key,
        )
        return result



import unittest
from datetime import date, datetime

from nanohttp import json, settings, context
from sqlalchemy import Unicode, Integer, Date, Float, ForeignKey, Boolean, DateTime
from sqlalchemy.orm import synonym
from sqlalchemy.ext.associationproxy import association_proxy

from restfulpy.controllers import JsonPatchControllerMixin, ModelRestController
from restfulpy.orm import commit, DeclarativeBase, Field, DBSession, composite, FilteringMixin, PaginationMixin, \
    OrderingMixin, relationship, ModifiedMixin, ActivationMixin
from restfulpy.testing import WebAppTestCase
from restfulpy.testing.helpers import MockupApplication


class FullName(object):  # pragma: no cover
    def __init__(self, first_name, last_name):
        self.first_name = first_name
        self.last_name = last_name

    def __composite_values__(self):
        return '%s %s' % (self.first_name, self.last_name)

    def __repr__(self):
        return "FullName(%s %s)" % (self.first_name, self.last_name)

    def __eq__(self, other):
        return isinstance(other, FullName) and \
               other.first_name == self.first_name and \
               other.last_name == self.last_name

    def __ne__(self, other):
        return not self.__eq__(other)


class Book(DeclarativeBase):
    __tablename__ = 'book'
    id = Field(Integer, primary_key=True)
    title = Field(Unicode(10), nullable=True)
    member_id = Field(Integer, ForeignKey("member.id"))


class Keyword(DeclarativeBase):
    __tablename__ = 'keyword'
    id = Field(Integer, primary_key=True)
    keyword = Field(Unicode(64))


class MemberKeywords(DeclarativeBase):
    __tablename__ = 'member_keywords'
    member_id = Field(Integer, ForeignKey("member.id"), primary_key=True)
    keyword_id = Field(Integer, ForeignKey("keyword.id"), primary_key=True)


class Member(ActivationMixin, ModifiedMixin, FilteringMixin, PaginationMixin, OrderingMixin, DeclarativeBase):
    __tablename__ = 'member'

    id = Field(Integer, primary_key=True)
    email = Field(Unicode(100), unique=True, index=True, json='email',
                  pattern=r'(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)', watermark='Email',
                  example="user@example.com", message='Invalid email address, please be accurate!', icon='email.svg')
    title = Field(Unicode(50), index=True, min_length=2, watermark='First Name')
    first_name = Field(Unicode(50), index=True, json='firstName', min_length=2, watermark='First Name')
    last_name = Field(Unicode(100), json='lastName', min_length=2, watermark='Last Name')
    phone = Field(
        Unicode(10), nullable=True, json='phone', min_length=10,
        pattern=r'\d{3}[-\.\s]??\d{3}[-\.\s]??\d{4}|\(\d{3}\)\s*\d{3}[-\.\s]??\d{4}|\d{3}[-\.\s]??\d{4}',
        watermark='Phone'
    )
    name = composite(FullName, first_name, last_name, readonly=True, json='fullName')
    _password = Field('password', Unicode(128), index=True, json='password', protected=True, min_length=6)
    birth = Field(Date)
    weight = Field(Float(asdecimal=True), default=50)
    _keywords = relationship('Keyword', secondary='member_keywords')
    keywords = association_proxy('_keywords', 'keyword', creator=lambda k: Keyword(keyword=k))
    visible = Field(Boolean, nullable=True)
    last_login_time = Field(DateTime, json='lastLoginTime')
    books = relationship('Book')

    def _set_password(self, password):
        self._password = 'hashed:%s' % password

    def _get_password(self):  # pragma: no cover
        return self._password

    password = synonym('_password', descriptor=property(_get_password, _set_password), info=dict(protected=True))


class Root(JsonPatchControllerMixin, ModelRestController):
    __model__ = Member

    @json
    @commit
    def post(self):
        m = Member()
        m.keywords.append('Hello')
        m.update_from_request()
        DBSession.add(m)
        return m

    @json
    @Member.expose
    def get(self, title: str=None):
        query = Member.query
        if title:
            return query.filter(Member.title == title).one_or_none()
        return query

    @json
    @Member.expose
    def me(self):
        return dict(
            title='me'
        )


class BaseModelTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', Root())
    __configuration__ = '''
    db:
      url: sqlite://    # In memory DB
      echo: false
    '''

    @classmethod
    def configure_app(cls):
        cls.application.configure(force=True)
        settings.merge(cls.__configuration__)
        settings.merge('''
        logging:
          loggers:
            default:
              level: debug
        ''')

    def test_update_from_request(self):
        resp, ___ = self.request(
            'ALL', 'POST', '/', params=dict(
                title='test',
                firstName='test',
                lastName='test',
                email='test@example.com',
                password='123456',
                birth='2001-01-01',
                weight=1.1,
                visible='false',
                lastLoginTime='2017-10-10T15:44:30.000',
                isActive=True
            ),
            doc=False
        )
        self.assertEqual(resp['title'], 'test')

        resp, ___ = self.request('ALL', 'GET', '/', query_string=dict(take=1), doc=False)
        self.assertEqual(len(resp), 1)
        self.assertEqual(resp[0]['title'], 'test')
        self.assertEqual(resp[0]['visible'], False)

        # 404
        self.request('ALL', 'GET', '/non-existence-user', doc=False, expected_status=404)

        # Plain dictionary
        resp, ___ = self.request('ALL', 'GET', '/me', doc=False)
        self.assertEqual(resp['title'], 'me')

        # 400 for sending relationship attribute
        self.request(
            'ALL', 'POST', '/', json=dict(
                title='test',
                firstName='test',
                lastName='test',
                email='test@example.com',
                password='123456',
                birth='2001-01-01',
                weight=1.1,
                visible='false',
                lastLoginTime='2017-10-10T15:44:30.000',
                isActive=True,
                books=[]
            ),
            doc=False,
            expected_status=400
        )

    def test_iter_columns(self):
        columns = {c.key: c for c in Member.iter_columns(relationships=False, synonyms=False, composites=False)}
        self.assertEqual(len(columns), 15)
        self.assertNotIn('name', columns)
        self.assertNotIn('password', columns)
        self.assertIn('_password', columns)

    def test_iter_json_columns(self):
        columns = {c.key: c for c in Member.iter_json_columns(
            include_readonly_columns=False, include_protected_columns=False)}
        self.assertEqual(len(columns), 13)
        self.assertNotIn('name', columns)
        self.assertNotIn('password', columns)
        self.assertNotIn('_password', columns)

    def test_metadata(self):
        resp, ___ = self.request('ALL', 'METADATA', '/', doc=False)

        self.assertIn('fields', resp)
        self.assertIn('name', resp)
        self.assertIn('primaryKeys', resp)
        self.assertIn('id', resp['primaryKeys'])
        self.assertEqual(resp['name'], 'Member')
        fields = resp['fields']
        self.assertIn('id', fields)
        self.assertIn('firstName', fields)
        self.assertEqual(fields['id']['primaryKey'], True)
        self.assertEqual(fields['email']['primaryKey'], False)
        self.assertEqual(fields['title']['primaryKey'], False)
        self.assertEqual(fields['title']['minLength'], 2)
        self.assertEqual(fields['title']['maxLength'], 50)
        self.assertEqual(fields['email']['pattern'], '(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)')

        self.assertEqual(fields['firstName']['name'], 'firstName')
        self.assertEqual(fields['firstName']['key'], 'first_name')

        self.assertEqual(fields['firstName']['type_'], 'str')
        self.assertEqual(fields['birth']['type_'], 'date')

        self.assertEqual(fields['weight']['default_'], 50)
        self.assertEqual(fields['visible']['optional'], True)

        self.assertEqual(fields['email']['message'], 'Invalid email address, please be accurate!')
        self.assertEqual(fields['email']['watermark'], 'Email')
        self.assertEqual(fields['email']['label'], 'Email')
        self.assertEqual(fields['email']['icon'], 'email.svg')
        self.assertEqual(fields['email']['example'], 'user@example.com')

    def test_datetime_format(self):
        # datetime allows contain microseconds
        self.request(
            'ALL', 'POST', '/',
            params=dict(
                title='test',
                firstName='test',
                lastName='test',
                email='test1@example.com',
                password='123456',
                birth='2001-01-01',
                weight=1.1,
                visible='false',
                lastLoginTime='2017-10-10T10:10:00.12313'
            ),
            doc=False
        )
        self.request(
            'ALL', 'POST', '/',
            params=dict(
                title='test',
                firstName='test',
                lastName='test',
                email='test2@example.com',
                password='123456',
                birth='2001-01-01',
                weight=1.1,
                visible='false',
                lastLoginTime='2017-10-10T10:10:00.'
            ),
            doc=False,
            expected_status=400
        )

        self.request(
            'ALL', 'POST', '/',
            params=dict(
                title='test',
                firstName='test',
                lastName='test',
                email='test3@example.com',
                password='123456',
                birth='2001-01-01',
                weight=1.1,
                visible='false',
                lastLoginTime='2017-10-10T10:10:00'
            ),
            doc=False
        )

        # Invalid month value
        self.request(
            'ALL', 'POST', '/',
            params=dict(
                title='test',
                firstName='test',
                lastName='test',
                email='test4@example.com',
                password='123456',
                birth='2001-01-01',
                weight=1.1,
                visible='false',
                lastLoginTime='2017-13-10T10:10:00'
            ),
            doc=False,
            expected_status=400
        )

        # Invalid datetime format
        self.request(
            'ALL', 'POST', '/',
            params=dict(
                title='test',
                firstName='test',
                lastName='test',
                email='test4@example.com',
                password='123456',
                birth='2001-01-01',
                weight=1.1,
                visible='false',
                lastLoginTime='InvalidDatetime'
            ),
            doc=False,
            expected_status=400
        )

        # datetime might not have ending Z
        resp, ___ = self.request(
            'ALL', 'POST', '/',
            params=dict(
                title='test',
                firstName='test',
                lastName='test',
                email='test5@example.com',
                password='123456',
                birth='2001-01-01',
                weight=1.1,
                visible='false',
                lastLoginTime='2017-10-10T10:10:00.4546'
            ),
            doc=False
        )
        self.assertEqual(resp['lastLoginTime'], '2017-10-10T10:10:00.004546Z')

        # datetime containing ending Z
        resp, ___ = self.request(
            'ALL', 'POST', '/', params=dict(
                title='test',
                firstName='test',
                lastName='test',
                email='test6@example.com',
                password='123456',
                birth='2001-01-01',
                weight=1.1,
                visible='false',
                lastLoginTime='2017-10-10T10:10:00.4546Z'
            ),
            doc=False
        )
        self.assertEqual(resp['lastLoginTime'], '2017-10-10T10:10:00.004546Z')

    def test_date_format(self):
        # iso date format
        resp, ___ = self.request(
            'ALL', 'POST', '/',
            params=dict(
                title='test',
                firstName='test',
                lastName='test',
                email='test11@example.com',
                password='123456',
                birth='2001-01-01',
                weight=1.1,
                visible='false',
                lastLoginTime='2017-10-10T10:10:00.4546Z'
            ),
            doc=False
        )
        self.assertEqual(resp['birth'], '2001-01-01')
        self.assertEqual(resp['lastLoginTime'], '2017-10-10T10:10:00.004546Z')

        # posix timestamp
        resp, ___ = self.request(
            'ALL', 'POST', '/',
            params=dict(
                title='test',
                firstName='test',
                lastName='test',
                email='test12@example.com',
                password='123456',
                birth='1513434403',
                weight=1.1,
                visible='false',
                lastLoginTime=datetime(2017, 10, 10, 10, 10, 0, 4546).timestamp()
            ),
            doc=False
        )
        self.assertEqual(resp['birth'], '2017-12-16')
        self.assertEqual(resp['lastLoginTime'], '2017-10-10T10:10:00.004546Z')

        # none iso date format
        self.request(
            'ALL', 'POST', '/',
            params=dict(
                title='test',
                firstName='test',
                lastName='test',
                email='test13@example.com',
                password='123456',
                birth='01-01-01',
                weight=1.1,
                visible='false',
                lastLoginTime='2017-10-10T10:10:00.4546Z'
            ),
            doc=False,
            expected_status=400
        )

        # none iso date format
        self.request(
            'ALL', 'POST', '/',
            params=dict(
                title='test',
                firstName='test',
                lastName='test',
                email='test14@example.com',
                password='123456',
                birth='2001/01/01',
                weight=1.1,
                visible='false',
                lastLoginTime='2017-10-10T10:10:00.4546Z'
            ),
            doc=False,
            expected_status=400
        )


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import functools

from mako.lookup import TemplateLookup
from nanohttp import settings, action


_lookup = None


def get_lookup():
    global _lookup
    if _lookup is None:
        _lookup = TemplateLookup(directories=settings.templates.directories)
    return _lookup


def render_template(func, template_name):

    @functools.wraps(func)
    def wrapper(*args, **kwargs):

        result = func(*args, **kwargs)
        if hasattr(result, 'to_dict'):
            result = result.to_dict()
        elif not isinstance(result, dict):
            raise ValueError('The result must be an instance of dict, not: %s' % type(result))

        template_ = get_lookup().get_template(template_name)
        return template_.render(**result)

    return wrapper


template = functools.partial(action, content_type='text/html', inner_decorator=render_template)
